static BLEUUID serviceUUID("0000fff0-0000-1000-8000-00805f9b34fb"); // The remote service we wish to connect to.
static BLEUUID    charUUID("0000fff4-0000-1000-8000-00805f9b34fb"); // The characteristic of the remote service we are interested in.

static BLERemoteCharacteristic* pRemoteCharacteristic;
static BLEAdvertisedDevice* myDevice;

static void notifyCallback (
    BLERemoteCharacteristic* pBLERemoteCharacteristic,
    uint8_t* pData,
    size_t length,
    bool isNotify) 
{
    for(uint8_t aaa = 0; aaa<length; aaa++)
        { DMMdata[aaa] = msb(pData[aaa]^ai_XOR[aaa]); }
}

class MyClientCallback : public BLEClientCallbacks 
{
    void onConnect(BLEClient* pclient) { }

    void onDisconnect(BLEClient* pclient) 
        {   connected = false;  }
};

bool connectToServer() 
{
    BLEClient*  pClient  = BLEDevice::createClient();
    pClient->setClientCallbacks(new MyClientCallback());
    pClient->connect(myDevice);  // if you pass BLEAdvertisedDevice instead of address, it will be recognized type of peer device address (public or private)
    pClient->setMTU(517); //set client to request maximum MTU from server (default is 23 otherwise)
      
    BLERemoteService* pRemoteService = pClient->getService(serviceUUID); // Obtain a reference to the service we are after in the remote BLE server.
    if (pRemoteService == nullptr)      //  Failed to find our service UUID:
    {
        pClient->disconnect();
        return false;
    }

    pRemoteCharacteristic = pRemoteService->getCharacteristic(charUUID);
    if (pRemoteCharacteristic == nullptr) 
    {
        Serial.print("Failed to find our characteristic UUID: ");
        Serial.println(charUUID.toString().c_str());
        pClient->disconnect();
        return false;
    }

    if (pRemoteCharacteristic->canRead())   {  }

    if (pRemoteCharacteristic->canNotify())
        {   pRemoteCharacteristic->registerForNotify(notifyCallback); }

    connected = true;
    return true;
}

class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks 
{
    void onResult(BLEAdvertisedDevice advertisedDevice) 
    {
        ////if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(serviceUUID)) 
        if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(serviceUUID) && (advertisedDevice.getAddress().toString() == cfg_addr))         
        {
            BLEDevice::getScan()->stop();
            myDevice = new BLEAdvertisedDevice(advertisedDevice);
            doConnect = true;
            doScan = true;
        }
  }
};


void BLE_init()
{
    BLEDevice::init("");
    BLEScan* pBLEScan = BLEDevice::getScan();
    pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
    pBLEScan->setInterval(1349);
    pBLEScan->setWindow(449);
    pBLEScan->setActiveScan(true);
    pBLEScan->start(5, false);
}
