/*
Verze pro dekodovani dat z BLE Aneng 9002

ESP32 Dev Module
4MB
Huge App (3MB no OTA/1MB SPIFFS)

v.1
*/

const char compile_date[] = __DATE__ ", " __TIME__;
  
#define SCPI_ARRAY_SYZE 8
#define SCPI_MAX_TOKENS 64
#define SCPI_MAX_COMMANDS 64
#define SCPI_BUFFER_LENGTH 128

#include "BLEDevice.h"
#include <EEPROM.h>
#include "Vrekrer_scpi_parser.h"

////#define LED 22
#define LED 2

SCPI_Parser scpi;

//String cfg_addr = "9c:0c:35:fe:f6:9f";
String cfg_addr = "00:00:00:00:00:00";
uint8_t cfg_auto = 0;
uint8_t cfg_id = 31;
uint8_t cfg_units = 1;
uint8_t cfg_baud = 9;   // 0 300, 1 600, 2 1200, 3 2400, 4 4800, 5 9600, 6 19200, 7 38400, 8 57600, 9 115200
uint8_t cfg_sp = 2;     // 0 - ";", 1 - ",", 2 - "; ", 3 = ", "
const String cfg_sep[] = { ";", ",", "; ", ", " } ;

const uint32_t baudrate[] = { 300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200 };

static boolean doConnect = false;
static boolean connected = false;
static boolean doScan = false;

int  DMMdata[11];
int  ai_XOR[] = {0x41,0x21,0x73,0x55,0xa2,0xc1,0x32,0x71,0x66,0xaa,0x3b,0xd0,0xe2};

String   s_SevSeg="";
String   s_DMM="no Value"; // Value
String   s_PRE="";         // Prefix M k _ m mikro n
String   s_UNIT="";        // 'C 'F V A Ohm Hz
String   s_OPT="";         // Optionen BT,HOLD,DC,AC

float dmm_value = 0.0;
String dmm_unit = "";

uint32_t tmr = 0;

void setup() 
{
    Serial.begin(115200);
    delay(200);
    

    EEPROM.begin(512);
    pinMode(LED, OUTPUT);
    digitalWrite(LED, HIGH);

    cfg_addr    = EEPROM.readString(0);
    cfg_auto    = EEPROM.read(20);
    cfg_id      = EEPROM.read(21);
    cfg_units   = EEPROM.read(22);
    cfg_baud    = EEPROM.read(23);
    cfg_sp      = EEPROM.read(24);

    if (cfg_auto == 0xff)
    {
        ////cfg_addr = "00:00:00:00:00:00";
        cfg_addr = "9c:0c:35:fe:f6:9f";
        cfg_auto = 1;
        cfg_id = 31;
        cfg_units = 1;
        cfg_baud = 9;
        cfg_sp = 3;
    }

    cfg_addr.toLowerCase();

    Serial.begin(baudrate[cfg_baud]);

    Serial.println("\r\n\r\nBLE interface for Aneng AN9002, S/N: "+cfg_addr+", "+String(compile_date));
    Serial.println("Type *HLP?<NL> for HELP.");

    BLE_init();
    doConnect = false;
    doScan = true;

    SCPI_init();
} 

void loop() 
{
    scpi.ProcessInput(Serial, "\n");

    if (doConnect == true) 
    {
        if (connectToServer()) 
        {   
            Serial.println("Aneng AN9002, S/N: "+cfg_addr+ " connected."); 
            Serial.println("Type *HLP?<NL> for HELP.");
            tmr = millis(); 
            doConnect = false; 
        } 
        else 
        {   
            Serial.println("Failed to connect.");  
            doConnect = true;  
        }        
    }
    else if (connected) 
    {   
        digitalWrite(LED, HIGH);
        decode_DMM(DMMdata);    

        if (cfg_auto > 0)
        {
            if (millis() > (tmr+250))
            {
                if (dmm_value >= 0)
                    {   Serial.print("+"); }

                if (abs(dmm_value) > 60)
                    {   Serial.print(dmm_value, 1); }
                else if (abs(dmm_value) > 6)
                    {   Serial.print(dmm_value, 2); }        
                else
                    {   Serial.print(dmm_value, 3); }

                Serial.print(cfg_sep[cfg_sp]);
                Serial.print(s_UNIT);
                Serial.print(cfg_sep[cfg_sp]);
                Serial.print(s_OPT);
                Serial.println(cfg_sep[cfg_sp]);   
                tmr = millis();          
            }
        }

        if (s_DMM == " 0.L ")
            {   dmm_value = 99.99E+99;   }
        else if (s_DMM == " 0L. ")
            {   dmm_value = 99.99E+99;  }
        else
            {   dmm_value = s_DMM.toFloat(); }
        
        delay(1);
        digitalWrite(LED, LOW);      
        delay(49);        
    } 
    else
    {           
        digitalWrite(LED, HIGH); 
        BLE_init();
    }
} 
