void SCPI_init()
{
    scpi.timeout = 20; //value in miliseconds. Default value = 10

    scpi.SetCommandTreeBase("");
        scpi.RegisterCommand(F("*IDN?"),           &SCPI_Identify);     //
        scpi.RegisterCommand(F("*HLP?"),           &SCPI_Help);         //
        scpi.RegisterCommand(F("*HLP"),            &SCPI_Help);         //
        scpi.RegisterCommand(F("*SAV?"),           &SCPI_GetSav); 
        scpi.RegisterCommand(F("*SAV"),            &SCPI_Sav);          //  

    scpi.SetCommandTreeBase(F("SYSTem"));
        scpi.RegisterCommand(F(":ADDRess?"),        &SCPI_AddrGet);      //
        scpi.RegisterCommand(F(":ADDRess"),         &SCPI_AddrSet);      //
        scpi.RegisterCommand(F(":BAUD?"),           &SCPI_BaudGet);   
        scpi.RegisterCommand(F(":BAUD"),            &SCPI_BaudSet);
        scpi.RegisterCommand(F(":AUTOsend?"),       &SCPI_AutoGet);         //         
        scpi.RegisterCommand(F(":AUTOsend"),        &SCPI_AutoSet);         //
        scpi.RegisterCommand(F(":UNITs?"),          &SCPI_UnitGet);     
        scpi.RegisterCommand(F(":UNITs"),           &SCPI_UnitSet);         
        scpi.RegisterCommand(F(":SEParator?"),      &SCPI_SepGet);            
        scpi.RegisterCommand(F(":SEParator"),       &SCPI_SepSet);         
        scpi.RegisterCommand(F(":ID?"),             &SCPI_IdGet);                   
        scpi.RegisterCommand(F(":ID"),              &SCPI_IdSet);         

    scpi.SetCommandTreeBase(F("MEASure"));
        scpi.RegisterCommand(F(":VOLTage:AC?"),     &SCPI_MeasVoltAC);      //
        scpi.RegisterCommand(F(":VOLTage:DC?"),     &SCPI_MeasVoltDC);      //
        scpi.RegisterCommand(F(":CURRent:DC?"),     &SCPI_MeasCurrDC);      //
        scpi.RegisterCommand(F(":CURRent:AC?"),     &SCPI_MeasCurrAC);      // 
        scpi.RegisterCommand(F(":RESistance?"),     &SCPI_MeasRes);         //
        scpi.RegisterCommand(F(":CAPacitance?"),    &SCPI_Cap);
        scpi.RegisterCommand(F(":TEMPerature:C?"),  &SCPI_MeasTempC);       // 
        scpi.RegisterCommand(F(":TEMPerature:F?"),  &SCPI_MeasTempF);       // 

        scpi.RegisterCommand(F(":DIOde?"),          &SCPI_Dio);
        scpi.RegisterCommand(F(":FREQuency?"),      &SCPI_Freq);
        scpi.RegisterCommand(F(":DUTYcycle?"),      &SCPI_Duty); 
}

//-------------------------------------------------------------------------------
void SCPI_Help(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println(F("--- HELP - list of all implemented SCPI commands ----------------------------"));
    interface.println(F("*IDN?                        / System Identify"));
    interface.println(F("*HLP?                        / This Help"));
    interface.println(F("*SAV?                        / Display All Saved System Settings"));  
    interface.println(F("*SAV                         / Save All System Settings"));  
    interface.println(F("SYSTem:ADDRess?                    / GET Client MAC address"));  
    interface.println(F("      :ADDRess [00:00:00:00:00:00] / SET Client MAC address"));
    interface.println(F("      :ID?                   / GET System ID"));       
    interface.println(F("      :ID [0..31]            / SET System ID"));
    interface.println(F("      :BAUD?                 / GET Communication Baud rate"));           
    interface.println(F("      :BAUD [0..9]           / SET Communication Baud rate"));
    interface.println(F("      :AUTOsend?             / GET autosend [0 none | 1 RAW]"));
    interface.println(F("      :AUTOsend [0..1]       / SET autosend [0 none | 1 RAW]"));
    interface.println(F("      :UNIT?                 / GET Display Unit [0 none | 1 display]"));      
    interface.println(F("      :UNIT [0..1]           / SET Display Unit [0 none | 1 display]"));
    interface.println(F("      :SEParator?            / GET Separator [0 \";\" | 1 \",\" | 2 \"; \" | 3 \", \" ]"));      
    interface.println(F("      :SEParator [0..3]      / GET Separator [0 \";\" | 1 \",\" | 2 \"; \" | 3 \", \" ]"));
    interface.println(F("MEASure:VOLTage:DC?          / GET Measure DC Voltage [V]"));
    interface.println(F("       :VOLTage:AC?          / GET Measure AC Voltage [V]"));  
    interface.println(F("       :CURRent:DC?          / GET Measure DC Current [A]"));
    interface.println(F("       :CURRent:AC?          / GET Measure AC Current [A]"));  
    interface.println(F("       :RESistance?          / GET Measure Resistance [Ohm]"));
    interface.println(F("       :CAPacitance?         / GET Measure Capacitance [F]"));    
    interface.println(F("       :TEMPerature:C?       / GET Measure Temperature [""\xC2\xB0""C]"));
    interface.println(F("       :TEMPerature:F?       / GET Measure Temperature [""\xC2\xB0""F]"));      
    interface.println(F("       :DIODe?               / GET Measure Diode [V]"));
    interface.println(F("       :FREQuency?           / GET Measure Frequency [Hz]"));
    interface.println(F("       :DUTYcycle?           / GET Measure Duty cycle [%]")); 
    interface.println(F("--- HELP --------------------------------------------------------------------"));
}

//-------------------------------------------------------------------------------
void SCPI_Identify(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.print("Aneng AN9002, S/N: "+cfg_addr);
    if (connected)
        { interface.println(" connected.");     }
    else
        { interface.println(" not connected."); }
}
//-------------------------------------------------------------------------------
void SCPI_AddrGet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println("ADDR: "+cfg_addr);
}
//-------------------------------------------------------------------------------
void SCPI_AddrSet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    if (parameters.Size() > 0) 
    {        
        cfg_addr = String(parameters[0]);
        cfg_addr.toLowerCase();
        interface.println("ADDR set to: "+cfg_addr);
    }      
}
//-------------------------------------------------------------------------------
void SCPI_BaudGet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println("BAUD: "+String(baudrate[cfg_baud]));
}
//-------------------------------------------------------------------------------
void SCPI_BaudSet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    if (parameters.Size() > 0) 
    {        
        cfg_baud = constrain(String(parameters[0]).toInt(), 0, 9);
        interface.println("BAUD set to: "+String(baudrate[cfg_baud]));
        Serial.begin(baudrate[cfg_baud]);
    }      
}
//-------------------------------------------------------------------------------
void SCPI_AutoGet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println("AUTO: "+String(cfg_auto));
}
//-------------------------------------------------------------------------------
void SCPI_AutoSet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    if (parameters.Size() > 0) 
    {        
        cfg_auto = constrain(String(parameters[0]).toInt(), 0, 1);
        interface.println("AUTO set to: "+String(cfg_auto));
    }
}
//-------------------------------------------------------------------------------
void SCPI_UnitGet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println("UNIT: "+String(cfg_units));
}
//-------------------------------------------------------------------------------
void SCPI_UnitSet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    if (parameters.Size() > 0) 
    {        
        cfg_units = constrain(String(parameters[0]).toInt(), 0, 1);
        interface.println("UNIT set to: "+String(cfg_units));
    }    
}
//-------------------------------------------------------------------------------
void SCPI_SepGet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println("SEParator: \""+cfg_sep[cfg_sp]+"\"");
}
//-------------------------------------------------------------------------------
void SCPI_SepSet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    if (parameters.Size() > 0) 
    {        
        cfg_sp = constrain(String(parameters[0]).toInt(), 0, 3);
        interface.println("SEParator set to: \""+cfg_sep[cfg_sp]+"\"");
    }      
}
//-------------------------------------------------------------------------------
void SCPI_IdGet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    interface.println("ID: "+String(cfg_id));
}
//-------------------------------------------------------------------------------
void SCPI_IdSet(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    if (parameters.Size() > 0) 
    {        
        cfg_id = constrain(String(parameters[0]).toInt(), 0, 31);
        interface.println("ID set to: "+String(cfg_id));
    }    
}
//-------------------------------------------------------------------------------
void SCPI_MeasVoltDC (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"V DC"; }

    if (connected) 
    {
        if (s_UNIT == "V DC")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else if (s_UNIT == "mV DC")
        {
            interface.print(sci(dmm_value*0.001, 5));
            interface.println(unit);
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_MeasVoltAC (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"V AC"; }

    if (connected) 
    {
        if (s_UNIT == "V AC")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else if (s_UNIT == "mV AC")
        {
            interface.print(sci(dmm_value*0.001, 5));
            interface.println(unit);
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_MeasCurrDC (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"A DC"; }

    if (connected) 
    {
        if (s_UNIT == "A DC")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else if (s_UNIT == "mA DC")
        {
            interface.print(sci(dmm_value*0.001, 5));
            interface.println(unit);
        }
        else if (s_UNIT == "uA DC")
        {
            interface.print(sci(dmm_value*0.000001, 5));
            interface.println(unit);
        }        
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_MeasCurrAC (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{    
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"A AC"; }

    if (connected) 
    {
        if (s_UNIT == "A AC")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else if (s_UNIT == "mA AC")
        {
            interface.print(sci(dmm_value*0.001, 5));
            interface.println(unit);
        }
        else if (s_UNIT == "uA AC")
        {
            interface.print(sci(dmm_value*0.000001, 5));
            interface.println(unit);
        }        
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_MeasRes (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"Ohm"; }

    if (connected) 
    {
        if (s_UNIT == "Ohm")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else if (s_UNIT == "kOhm")
        {
            interface.print(sci(dmm_value*1000.0, 5));
            interface.println(unit);
        }
        else if (s_UNIT == "MOhm")
        {
            interface.print(sci(dmm_value*1000000.0, 5));
            interface.println(unit);
        }        
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_MeasTempC (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"\xC2\xB0""C"; }

    if (connected) 
    {
        if (s_UNIT == "\xC2\xB0""C")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_MeasTempF (SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"\xC2\xB0""F"; }

    if (connected) 
    {
        if (s_UNIT == "\xC2\xB0""F")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_Dio(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"V"; }

    if (connected) 
    {
        if (s_UNIT == "V D")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_Freq(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"Hz"; }

    if (connected) 
    {
        if (s_UNIT == "Hz")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_Duty(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
   String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"%"; }

    if (connected) 
    {
        if (s_UNIT == "%")
        {
            interface.print(sci(dmm_value, 5));  
            interface.println(unit);          
        }
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_Cap(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    String unit = "";
    if (cfg_units == 1)
        { unit = cfg_sep[cfg_sp]+"F"; }

    if (connected) 
    {
        if (s_UNIT == "nF")
        {
            interface.print(sci(dmm_value*1.0E-9, 5));  
            interface.println(unit);          
        }
        else if (s_UNIT == "uF")
        {
            interface.print(sci(dmm_value*1.0E-6, 5)); 
            interface.println(unit);
        }
        else if (s_UNIT == "mF")
        {
            interface.print(sci(dmm_value*1.0E-3, 5)); 
            interface.println(unit);
        }        
        else
            {   interface.println("N/A");   }
    }
    else
        {   interface.println("N/A");       }
}
//-------------------------------------------------------------------------------
void SCPI_GetSav(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    cfg_addr  = EEPROM.readString(0);
    cfg_auto  = EEPROM.read(20);
    cfg_id    = EEPROM.read(21);
    cfg_units = EEPROM.read(22);
    cfg_baud  = EEPROM.read(23);  
    cfg_sp    = EEPROM.read(24);

    interface.println("Saved parameters:");
    interface.println("Addr:      "+String(cfg_addr));
    interface.println("Auto:      "+String(cfg_auto));
    interface.println("ID:        "+String(cfg_id));
    interface.println("Unit:      "+String(cfg_units));
    interface.println("Separator: \""+String(cfg_sep[cfg_sp])+"\"");
    interface.println("Baud:      "+String(cfg_baud)+" : "+String(baudrate[cfg_baud]));        
}

//-------------------------------------------------------------------------------
void SCPI_Sav(SCPI_C commands, SCPI_P parameters, Stream& interface) 
{
    EEPROM.writeString(0, cfg_addr);
    EEPROM.write(20, cfg_auto);
    EEPROM.write(21, cfg_id);
    EEPROM.write(22, cfg_units);
    EEPROM.write(23, cfg_baud);
    EEPROM.write(24, cfg_sp);
    EEPROM.commit();

    cfg_addr  = EEPROM.readString(0);
    cfg_auto  = EEPROM.read(20);
    cfg_id    = EEPROM.read(21);
    cfg_units = EEPROM.read(22);
    cfg_baud  = EEPROM.read(23);  
    cfg_sp    = EEPROM.read(24);

    interface.println("Parameters SAVED");
    interface.println("Addr:      "+String(cfg_addr));
    interface.println("Auto:      "+String(cfg_auto));
    interface.println("ID:        "+String(cfg_id));
    interface.println("Unit:      "+String(cfg_units));
    interface.println("Separator: \""+String(cfg_sep[cfg_sp])+"\"");
    interface.println("Baud:      "+String(cfg_baud)+" : "+String(baudrate[cfg_baud]));   
    interface.println(); 
    interface.println("Restart with saved parameters..."); 
    interface.println();
    delay(3000);
    ESP.restart();     
}

//===============================================================================
char __printbuffer[66];
//  typical buffer size for 8 byte double is 22 bytes
//  15 bytes mantissa, sign dot E-xxx
//  em = exponentMultiple.
char * scieng(double value, uint8_t decimals, uint8_t em)
{
    
    char *  buffer   = __printbuffer;
    int     exponent = 0;
    uint8_t pos = 0;
    double  e1 = 10;
    double  e2 = 0.1;

    for (uint8_t i = 1; i < em; i++)  //  scale to multiples of em.
    {
        e1 *= 10;
        e2 *= 0.1;
    }

    if (isnan(value))
    {
        strcpy(buffer, "nan");
        return buffer;
    }

    if (isinf(value))
    {
        if (value < 0) strcpy(buffer, "-inf");
        strcpy(buffer, "inf");
        return buffer;
    }

    if (value < 0.0)      //  Handle negative numbers
    {
        buffer[pos++] = '-';
        value = -value;
    }
    else /////////////////////////////// to je moje
    {
        buffer[pos++] = '+';
    }

    while (value >= e1)
    {
        value *= e2;
        exponent += em;
    }

    while (value < 1 && value != 0.0)
    {
        value *= e1;
        exponent -= em;
    }

    double rounding = 0.5;        //  Round correctly so that print(1.999, 2) prints as "2.00"
    for (uint8_t i = 0; i < decimals; ++i)
    {
        rounding *= 0.1;
    }

    value += rounding;
    if (value >= e1)
    {
        exponent += em;
        value *= e2;
    }

    uint32_t d = (uint32_t)value;     //  Split whole part and remainder
    double remainder = value - d;

    itoa(d, &buffer[pos], 10);    //  print whole part
    pos = strlen(buffer);

    if (decimals > 0)         //  print remainder part
    {
        buffer[pos++] = '.';    //  decimal point
    }

    while (decimals-- > 0)    //  Extract decimals from the remainder one at a time to prevent missing leading zero's
    {
        remainder *= 10;
        d = (uint8_t)remainder;
        buffer[pos++] = d + '0';
        remainder -= d;
    }

    buffer[pos++] = 'E';  //  print exponent
    if (exponent < 0)
    {
        buffer[pos++] = '-';
        exponent = -exponent;
    }
    else buffer[pos++] = '+';

    if (exponent < 10) buffer[pos++] = '0';
    itoa(exponent, &buffer[pos], 10);
    return buffer;
}

char * sci(double value, uint8_t decimals)
{
    return scieng(value, decimals, 1);
}

size_t sci(Stream &str, double value, uint8_t decimals)
{
    return str.print(sci(value, decimals));
}